@@ -7,6 +7,7 @@ Cameron Gilchrist
 """
 
 import logging
+import uuid
 
 from collections import defaultdict, OrderedDict
 from itertools import combinations, product
@@ -20,7 +21,7 @@ from Bio import Align
 from Bio.Align import substitution_matrices
 
 from clinker.formatters import format_alignment, format_globaligner
-from clinker.classes import Cluster
+from clinker.classes import Serializer, Cluster, Locus, Gene, load_child, load_children
 
 
 LOG = logging.getLogger(__name__)
@@ -137,7 +138,7 @@ def compute_identity(alignment):
     return matches / length, (matches + similar) / length
 
 
-class Globaligner:
+class Globaligner(Serializer):
     """Performs and stores alignments.
 
     Parameters:
@@ -158,18 +159,109 @@ class Globaligner:
     }
 
     def __init__(self, aligner_config=None):
-        self.alignments = []
-        self.aligner = Align.PairwiseAligner()
-        self.clusters = OrderedDict()
-
+        # Lookup dictionaries
+        self._genes = {}
+        self._loci = {}
+        self._links = {}
         self._alignment_indices = defaultdict(dict)
         self._cluster_names = defaultdict(dict)
 
+        self.alignments = {}
+        self.aligner = Align.PairwiseAligner()
+        self.clusters = OrderedDict()
+
         if aligner_config:
             self.configure_aligner(**aligner_config)
         else:
             self.configure_aligner(**self.aligner_default)
 
+    def to_dict(self):
+        """Serialises the Globaligner instance to dict.
+
+        Schema:
+            {
+                order: [],
+                clusters: {},
+                loci: {},
+                genes: {},
+                alignments: [],
+                links: {},
+            }
+
+        Where each child dictionary holds serialised Python objects keyed
+        by their UIDs. When from_dict() is called, they are used to store
+        real references between objects (e.g. Link query/target attributes
+        are Gene objects).
+        """
+        return {
+            "order": list(self.clusters),
+            "clusters": {
+                uid: cluster.to_dict(uids_only=True)
+                for uid, cluster in self.clusters.items()
+            },
+            "loci": {
+                uid: locus.to_dict(uids_only=True)
+                for uid, locus in self._loci.items()
+            },
+            "genes": {
+                uid: gene.to_dict()
+                for uid, gene in self._genes.items()
+            },
+            "alignments": {
+                uid: alignment.to_dict(uids_only=True)
+                for uid, alignment in self.alignments.items()
+            },
+            "links": {
+                uid: link.to_dict()
+                for uid, link in self._links.items()
+            },
+        }
+
+    @classmethod
+    def from_dict(cls, d):
+        """Loads a Globaligner instance from dict generated by to_dict().
+
+        First, loads all clinker objects back into memory (cluster, locus, gene)
+        and restores their hierarchical structure. Alignments are restored in the same way.
+        Finally, rebuilds lookup dictionaries used by the Globaligner class.
+        """
+        ga = Globaligner()
+
+        # Reconstruct cluster -> locus -> gene hierarchy
+        for cluster_uid in d["order"]:
+            cluster = Cluster.from_dict(d["clusters"][cluster_uid])
+            for locus_idx, locus_uid in enumerate(cluster.loci):
+                locus = Locus.from_dict(d["loci"][locus_uid])
+                for gene_idx, gene_uid in enumerate(locus.genes):
+                    gene = Gene.from_dict(d["genes"][gene_uid])
+                    locus.genes[gene_idx] = gene
+                    ga._genes[gene_uid] = gene
+                cluster.loci[locus_idx] = locus
+                ga._loci[locus_uid] = locus
+            ga.clusters[cluster_uid] = cluster
+
+        # Reconstruct cluster alignments
+        for alignment_uid, alignment in d["alignments"].items():
+            aln = Alignment.from_dict(alignment)
+            aln.query = ga.clusters[aln.query]
+            aln.target = ga.clusters[aln.target]
+
+            # Form Link objects and add to lookup dict
+            for idx, uid in enumerate(aln.links):
+                link = Link.from_dict(d["links"][uid])
+                link.query = ga._genes[link.query.uid]
+                link.target = ga._genes[link.target.uid]
+                aln.links[idx] = link
+                ga._links[uid] = link
+
+            # Update Alignment object lookup dictionaries
+            ga.alignments[aln.uid] = aln
+            ga._alignment_indices[aln.query.uid][aln.target.uid] = aln.uid
+            ga._alignment_indices[aln.target.uid][aln.query.uid] = aln.uid
+            ga._cluster_names[aln.uid] = (aln.query.uid, aln.target.uid)
+
+        return ga
+
     def __str__(self):
         """Print all alignments currently stored in the instance."""
         return self.format()
@@ -189,24 +281,32 @@ class Globaligner:
             link_headers=link_headers,
         )
 
-    def to_data(self, i=0.5, method="ward"):
+    def to_data(self, i=0.5, method="ward", use_file_order=False):
         """Formats Globaligner as plottable data set.
 
         Assign unique indices to all clusters, loci, genes
         """
-        clusters = list(self.clusters.values())
+        clusters = [cluster.to_dict() for cluster in self.clusters.values()]
         return {
-            "clusters": [
-                clusters[i].to_dict()
-                for i in self.order(i=i, method=method)
+            "clusters": clusters if use_file_order else [
+                clusters[i] for i in self.order(i=i, method=method)
             ],
             "links": [
                 link.to_dict()
-                for alignment in self.alignments
+                for alignment in self.alignments.values()
                 for link in alignment.links
             ],
         }
 
+    @property
+    def gene_labels(self):
+        labels = set()
+        for cluster in self.clusters.values():
+            for locus in cluster.loci:
+                for gene in locus.genes:
+                    labels.update(gene.names)
+        return labels
+
     def add_clusters(self, *clusters):
         """Adds new Cluster object/s to the Globaligner.
 
@@ -216,8 +316,11 @@ class Globaligner:
         for cluster in clusters:
             if not isinstance(cluster, Cluster):
                 raise NotImplementedError("Expected Cluster object")
-            if cluster.name not in self.clusters:
-                self.clusters[cluster.name] = cluster
+            self.clusters[cluster.uid] = cluster
+            for locus in cluster.loci:
+                self._loci[locus.uid] = locus
+                for gene in locus.genes:
+                    self._genes[gene.uid] = gene
 
     def align_clusters(self, one, two, cutoff=0.3):
         """Constructs a cluster alignment using aligner in the Globaligner."""
@@ -234,7 +337,8 @@ class Globaligner:
     def align_stored_clusters(self, cutoff=0.3):
         """Aligns clusters stored in the Globaligner."""
         for one, two in combinations(self.clusters.values(), 2):
-            if self._alignment_indices[one.name].get(two.name):
+            if self._alignment_indices[one.uid].get(two.uid):
+                LOG.debug("Skipping %s vs %s", one.name, two.name)
                 continue
             LOG.info("%s vs %s", one.name, two.name)
             alignment = self.align_clusters(one, two, cutoff)
@@ -261,15 +365,7 @@ class Globaligner:
         """Returns a printout of the current PairwiseAligner object settings."""
         return str(self.aligner)
 
-    def form_alignment_string(self, index):
-        """Return a string representation of a stored Alignment."""
-        one, two = self._cluster_names[index]
-        header = f"{one} vs {two}"
-        separator = "-" * len(header)
-        alignment = self.alignments[index]
-        return f"{header}\n{separator}\n{alignment}"
-
-    def add_alignment(self, alignment):
+    def add_alignment(self, alignment, overwrite=False):
         """Adds a new cluster alignment to the Globaligner.
 
         self._alignment_indices allows for Alignment indices to be
@@ -284,17 +380,21 @@ class Globaligner:
         self.add_clusters(q, t)
 
         # Overwrite previous alignment between these clusters if one exists
-        index = self._alignment_indices[q.name].get(t.name)
-        if index:
-            self.alignments[index] = alignment
-        else:
-            index = len(self.alignments)
-            self.alignments.append(alignment)
+        previous = self._alignment_indices[q.uid].get(t.uid)
+
+        # Clear up any old links, add the new ones
+        if previous:
+            previous = self.alignments.pop(previous)
+            for link in previous.links:
+                self._links.pop(link.uid)
+        for link in alignment.links:
+            self._links[link.uid] = link
 
         # Update mapping dictionaries and save Alignment
-        self._alignment_indices[q.name][t.name] = index
-        self._alignment_indices[t.name][q.name] = index
-        self._cluster_names[index] = (q.name, t.name)
+        self.alignments[alignment.uid] = alignment
+        self._alignment_indices[q.uid][t.uid] = alignment.uid
+        self._alignment_indices[t.uid][q.uid] = alignment.uid
+        self._cluster_names[alignment.uid] = (q.uid, t.uid)
 
     def get_alignment(self, one, two):
         """Retrieves an Alignment corresponding to two Cluster objects.
@@ -305,10 +405,8 @@ class Globaligner:
         Returns:
             Alignment object for the specified Clusters
         """
-        index = self._alignment_indices[one].get(two)
-        if index is None:
-            raise KeyError(f"No alignment stored for {one} and {two}")
-        return self.alignments[index]
+        uid = self._alignment_indices[one][two]
+        return self.alignments[uid]
 
     def synteny(self, one, two, i=0.5):
         """Calculates a synteny score between two clusters.
@@ -324,15 +422,15 @@ class Globaligner:
         identity value of homologues in each cluster.
         """
         alignment = self.get_alignment(one, two)
+        if not alignment.links:
+            return 0
         homology = sum(link.identity for link in alignment.links)
-
         assign_groups(alignment.links)
         one_cluster = self.clusters[one]
         two_cluster = self.clusters[two]
         one_pairs = get_pairs(one_cluster)
         two_pairs = get_pairs(two_cluster)
         contiguity = compare_pairs(one_pairs, two_pairs)
-
         return homology + i * contiguity
 
     def matrix(self, i=0.5, normalise=False, as_distance=False):
@@ -353,7 +451,10 @@ class Globaligner:
                     continue
                 matrix[i, j] = self.synteny(one, two, i=i)
         if normalise:
-            matrix /= matrix.max()
+            # Explicitly check for zero so empty alignments don't cause issues
+            max_value = matrix.max()
+            if max_value > 0:
+                matrix /= max_value
         if as_distance:
             maximum = 1 if normalise else matrix.max()
             matrix = maximum - matrix
@@ -374,18 +475,35 @@ class Globaligner:
         return hierarchy.leaves_list(linkage)[::-1]
 
 
-class Alignment:
+class Alignment(Serializer):
     """An alignment between two gene clusters.
 
     Attributes:
         links (list): list of Gene-Gene 'links' (i.e. alignments)
     """
 
-    def __init__(self, query=None, target=None, links=None):
+    def __init__(self, uid=None, query=None, target=None, links=None):
+        self.uid = uid if uid else str(uuid.uuid4())
         self.query = query
         self.target = target
         self.links = links if links else []
 
+    def to_dict(self, uids_only=False):
+        return {
+            "uid": self.uid,
+            "query": self.query.uid if uids_only else self.query.to_dict(),
+            "target": self.target.uid if uids_only else self.target.to_dict(),
+            "links": [link.uid if uids_only else link.to_dict() for link in self.links]
+        }
+
+    @classmethod
+    def from_dict(cls, d):
+        return cls(
+            query=load_child(d["query"], Cluster),
+            target=load_child(d["target"], Cluster),
+            links=load_children(d["links"], Link),
+        )
+
     def __str__(self):
         return self.format()
 
@@ -419,14 +537,20 @@ class Alignment:
 
     def add_link(self, query, target, identity, similarity):
         """Instantiate a new Link from a Gene alignment and save."""
-        link = Link(query, target, identity=identity, similarity=similarity)
+        link = Link(
+            query=query,
+            target=target,
+            identity=identity,
+            similarity=similarity
+        )
         self.links.append(link)
 
 
-class Link:
+class Link(Serializer):
     """An alignment link between two Gene objects."""
 
-    def __init__(self, query, target, identity=None, similarity=None):
+    def __init__(self, uid=None, query=None, target=None, identity=None, similarity=None):
+        self.uid = uid if uid else str(uuid.uuid4())
         self.query = query
         self.target = target
         self.identity = identity
@@ -438,10 +562,17 @@ class Link:
     def values(self):
         return [self.query.name, self.target.name, self.identity, self.similarity]
 
-    def to_dict(self):
+    def to_dict(self, uids_only=False):
         return {
-            "query": dict(uid=self.query.uid, name=self.query.name),
-            "target": dict(uid=self.target.uid, name=self.target.name),
+            "uid": self.uid,
+            "query": self.query.uid if uids_only else self.query.to_dict(),
+            "target": self.target.uid if uids_only else self.target.to_dict(),
             "identity": self.identity,
             "similarity": self.similarity,
         }
+
+    @classmethod
+    def from_dict(cls, d):
+        d["query"] = load_child(d["query"], Gene)
+        d["target"] = load_child(d["target"], Gene)
+        return cls(**d)
