@@ -6,8 +6,9 @@ Module to store classes used across the package.
 Cameron Gilchrist
 """
 
-import itertools
+import json
 import warnings
+import uuid
 
 from pathlib import Path
 
@@ -25,6 +26,25 @@ def find_qualifier(valid_values, qualifiers):
     return None
 
 
+def subdict(d, keys):
+    """Creates a sub-dictionary of a parent dictionary given set of keys."""
+    funcs = ("lower", "upper", "title")
+    sub = {}
+    for key, value in d.items():
+        if key in keys or any(getattr(key, func) in keys for func in funcs):
+            if isinstance(value, list):
+                sub[key] = value[0]
+            else:
+                sub[key] = value
+    return sub
+
+
+def get_value(d, keys):
+    for key in keys:
+        if key in d:
+            return d[key]
+
+
 def parse_genbank(path):
     path = Path(path)
     with path.open() as fp:
@@ -50,7 +70,60 @@ def parse_files(paths):
     return [parse_genbank(path) for path in paths]
 
 
-class Cluster:
+def get_children(children, uids_only=False):
+    return [
+        child.uid
+        if uids_only
+        else child.to_dict()
+        for child in children
+    ]
+
+
+def load_child(child, thing):
+    if not hasattr(thing, "from_dict"):
+        raise NotImplementedError
+    return child if isinstance(child, str) else thing.from_dict(child)
+
+
+def load_children(children, thing):
+    return [load_child(child, thing) for child in children]
+
+
+class Serializer:
+    """JSON serialisation mixin class.
+
+    Classes that inherit from this class should implement `to_dict` and
+    `from_dict` methods.
+    """
+
+    def to_dict(self):
+        """Serialises class to dict."""
+        raise NotImplementedError
+
+    @classmethod
+    def from_dict(self, d):
+        """Loads class from dict."""
+        raise NotImplementedError
+
+    def to_json(self, fp=None, **kwargs):
+        """Serialises class to JSON."""
+        d = self.to_dict()
+        if fp:
+            json.dump(d, fp, **kwargs)
+        else:
+            return json.dumps(d, **kwargs)
+
+    @classmethod
+    def from_json(cls, js):
+        """Instantiates class from JSON handle."""
+        if isinstance(js, str):
+            d = json.loads(js)
+        else:
+            d = json.load(js)
+        return cls.from_dict(d)
+
+
+class Cluster(Serializer):
     """The Cluster class stores Proteins
 
     Attributes:
@@ -59,20 +132,26 @@ class Cluster:
           to Gene objects in self.genes
     """
 
-    id_iter = itertools.count()
-
     def __init__(self, name, loci, uid=None):
-        self.uid = uid if uid else next(Cluster.id_iter)
+        self.uid = uid if uid else str(uuid.uuid4())
         self.name = name
         self.loci = loci
 
-    def to_dict(self):
+    def to_dict(self, uids_only=False):
         return {
             "uid": self.uid,
             "name": self.name,
-            "loci": [locus.to_dict() for locus in self.loci]
+            "loci": get_children(self.loci, uids_only=uids_only)
         }
 
+    @classmethod
+    def from_dict(cls, d):
+        return cls(
+            d["name"],
+            load_children(d["loci"], Locus),
+            uid=d.get("uid")
+        )
+
     @classmethod
     def from_seqrecords(cls, *args, name=None):
         """Instantiates a Cluster from BioPython SeqRecord/s.
@@ -91,13 +170,11 @@ class Cluster:
                 return gene
 
 
-class Locus:
+class Locus(Serializer):
     """A cluster locus."""
 
-    id_iter = itertools.count()
-
     def __init__(self, name, genes, start=None, end=None, uid=None):
-        self.uid = uid if uid else next(Locus.id_iter)
+        self.uid = uid if uid else str(uuid.uuid4())
         self.name = name
         self.genes = genes
         self.start = start
@@ -106,15 +183,25 @@ class Locus:
     def __str__(self):
         return f"{self.name}:{self.start}-{self.end}"
 
-    def to_dict(self):
+    def to_dict(self, uids_only=False):
         return {
             "uid": self.uid,
             "name": self.name,
             "start": self.start,
             "end": self.end,
-            "genes": [gene.to_dict() for gene in self.genes],
+            "genes": get_children(self.genes, uids_only=uids_only),
         }
 
+    @classmethod
+    def from_dict(cls, d):
+        return cls(
+            d["name"],
+            load_children(d["genes"], Gene),
+            start=d["start"],
+            end=d["end"],
+            uid=d.get("uid"),
+        )
+
     @classmethod
     def from_seqrecord(cls, record):
         """Builds a new Locus from a BioPython SeqRecord."""
@@ -134,45 +221,45 @@ class Locus:
                 return gene
 
 
-class Gene:
+class Gene(Serializer):
     """Location, annotation and attachment points for drawing links."""
 
-    id_iter = itertools.count()
-
     def __init__(
         self,
-        name=None,
+        uid=None,
+        label=None,
+        names=None,
         start=None,
         end=None,
         strand=None,
-        function=None,
         sequence=None,
         translation=None,
-        uid=None,
     ):
-        self.uid = uid if uid else next(Gene.id_iter)
-        self.name = name
+        self.uid = uid if uid else str(uuid.uuid4())
+        self.label = label if label else self.uid
+        self.names = names if names else {}
         self.start = start
         self.end = end
         self.strand = strand
         self.sequence = sequence
         self.translation = translation
-        self.smcog = {}
-        self.function = function if function else 'Hypothetical'
 
     def to_dict(self):
         return {
             "uid": self.uid,
-            "name": self.name,
+            "label": self.label,
+            "names": self.names,
             "start": self.start,
             "end": self.end,
             "strand": self.strand,
             "sequence": self.sequence,
             "translation": self.translation,
-            "smcog": self.smcog,
-            "function": self.function,
         }
 
+    @classmethod
+    def from_dict(cls, d):
+        return cls(**d)
+
     @property
     def ordered_smcogs(self):
         """Returns an ordered list of smCOG hits."""
@@ -189,21 +276,16 @@ class Gene:
         """
         if "pseudo" in feature.qualifiers:
             return
-        name = find_qualifier(['protein_id', 'locus_tag', 'ID'], feature.qualifiers)
-        if not name:
-            raise ValueError(
-                "Could not determine a valid identifier"
-                f" from a CDS SeqFeature in {record.id}"
-            )
+        tags = ("protein_id", "locus_tag", "id", "gene", "label", "name")
+        qualifiers = {
+            k: v[0] if isinstance(v, list) else v
+            for k, v in feature.qualifiers.items()
+        }
         sequence = feature.extract(record.seq)
-        translation = find_qualifier(
-            ["translation"],
-            feature.qualifiers,
-        )
-        if not translation:
-            translation = sequence.translate()
+        translation = qualifiers.pop("translation") or sequence.translate()
         return cls(
-            name=name,
+            names=qualifiers,
+            label=get_value(qualifiers, tags),
             sequence=str(sequence),
             translation=str(translation),
             start=int(feature.location.start),
