@@ -26,6 +26,7 @@ LOG = logging.getLogger(__name__)
 
 def clinker(
     files,
+    session=None,
     identity=0.3,
     delimiter=None,
     decimals=2,
@@ -35,50 +36,77 @@ def clinker(
     no_align=False,
     hide_link_headers=False,
     hide_alignment_headers=False,
+    use_file_order=False,
+    json_indent=None,
 ):
     """Entry point for running the script."""
     LOG.info("Starting clinker")
 
-    # Check output file before doing anything else
-    if output and Path(output).exists() and not force:
-        LOG.error(f"File {output} already exists but --force not specified, exiting")
-        return
+    load_session = session and Path(session).exists()
 
-    # Parse files, generate objects
-    paths = find_files(files)
-    LOG.info("Parsing GenBank files: %s", paths)
-    clusters = parse_files(paths)
+    if load_session:
+        LOG.info("Loading session from: %s", session)
+        with open(session) as fp:
+            globaligner = align.Globaligner.from_json(fp)
+        if files:
+            paths = find_files(files)
+            LOG.info("Parsing GenBank files: %s", paths)
+            clusters = parse_files(paths)
 
-    # Align all clusters
-    if no_align:
-        globaligner = align.Globaligner()
-        globaligner.add_clusters(*clusters)
-    elif len(clusters) == 1:
-        globaligner = align.align_clusters(clusters[0])
+            LOG.info("Adding clusters to loaded session and aligning")
+            globaligner.add_clusters(*clusters)
+            globaligner.align_stored_clusters(cutoff=identity)
+            load_session = False
     else:
-        LOG.info("Starting cluster alignments")
-        globaligner = align.align_clusters(*clusters, cutoff=identity)
+        # Parse files, generate objects
+        paths = find_files(files)
+        LOG.info("Parsing GenBank files: %s", paths)
+        clusters = parse_files(paths)
 
-        LOG.info("Generating results summary...")
-        summary = globaligner.format(
-            delimiter=delimiter,
-            decimals=decimals,
-            link_headers=not hide_link_headers,
-            alignment_headers=not hide_alignment_headers,
-        )
-        if output:
-            LOG.info(f"Writing alignments to {output}")
+        # Align all clusters
+        if no_align:
+            globaligner = align.Globaligner()
+            globaligner.add_clusters(*clusters)
+        elif len(clusters) == 1:
+            globaligner = align.align_clusters(clusters[0])
+        else:
+            LOG.info("Starting cluster alignments")
+            globaligner = align.align_clusters(*clusters, cutoff=identity)
+
+    LOG.info("Generating results summary...")
+    summary = globaligner.format(
+        delimiter=delimiter,
+        decimals=decimals,
+        link_headers=not hide_link_headers,
+        alignment_headers=not hide_alignment_headers,
+    )
+
+    if output:
+        if (output and Path(output).exists() and not force):
+            print(summary)
+            LOG.warn("File %s already exists but --force was not specified", output)
+        else:
+            LOG.info("Writing alignments to: %s", output)
             with open(output, "w") as fp:
                 fp.write(summary)
-        else:
-            print(summary)
+    else:
+        print(summary)
+
+    if session and not load_session:
+        LOG.info("Saving session to: %s", session)
+        with open(session, "w") as fp:
+            globaligner.to_json(fp, indent=json_indent)
 
     # Generate the SVG
     if plot:
         LOG.info("Building clustermap.js visualisation")
         if isinstance(plot, str):
             LOG.info("Writing to: %s", plot)
-        plot_clusters(globaligner, output=None if plot is True else plot)
+        plot_clusters(
+            globaligner,
+            output=None if plot is True else plot,
+            use_file_order=use_file_order,
+        )
 
     LOG.info("Done!")
     return globaligner
@@ -118,6 +146,8 @@ def get_parser():
     )
 
     output = parser.add_argument_group("Output options")
+    output.add_argument("-s", "--session", help="Path to clinker session")
+    output.add_argument("-ji", "--json_indent", type=int, help="Number of spaces to indent JSON")
     output.add_argument("-f", "--force", help="Overwrite previous output file", action="store_true")
     output.add_argument("-o", "--output", help="Save alignments to file")
     output.add_argument(
@@ -145,6 +175,14 @@ def get_parser():
         action="store_true",
     )
 
+    viz = parser.add_argument_group("Visualisation options")
+    viz.add_argument(
+        "-ufo",
+        "--use_file_order",
+        action="store_true",
+        help="Display clusters in order of input files"
+    )
+
     return parser
 
 
@@ -153,6 +191,8 @@ def main():
     args = parser.parse_args()
     clinker(
         args.files,
+        session=args.session,
+        json_indent=args.json_indent,
         identity=args.identity,
         delimiter=args.delimiter,
         decimals=args.decimals,
@@ -162,6 +202,7 @@ def main():
         no_align=args.no_align,
         hide_link_headers=args.hide_link_headers,
         hide_alignment_headers=args.hide_aln_headers,
+        use_file_order=args.use_file_order,
     )
 
 
